/*
 * Class:    RowSet
 *
 */


import java.lang.*;
import java.math.BigDecimal;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.text.DecimalFormat;
import java.util.Vector;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;
import java.util.Calendar;
import org.apache.axis.encoding.Base64;
import org.apache.xerces.dom.DocumentImpl;
import org.w3c.dom.*;

/**
 * RowSet Class for used for serializing/deserializing SOAP parameters that represent
 * Progress Dynamic TEMP-TABLES.
 *
 */
public class RowSet
{
    private Vector m_rows;
    private int m_rowNum;
    private Vector m_currentRow;
    private ColumnMetaData[] m_schema;
    private DocumentImpl m_dom;
    private SimpleDateFormat m_sdfDate;
    private SimpleDateFormat m_sdfDateTime;
    /*
     * Constructors...
     */

    /**
    * <!--RowSet() -->
    * <p>The default class constructor.
    * </p>
    * <br>
    * <br>
    * @return   void
    * <br>
    * @exception    Exception
    */
    public RowSet()
    {
      m_rows = new Vector();
      m_rowNum = 0;
      m_currentRow = null;
      m_schema = null;
      m_dom = new DocumentImpl();
      m_sdfDate = new SimpleDateFormat("yyyy-MM-dd");
      m_sdfDateTime = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssz");
    }


    /**
     * next method used by to traverse the rows of the result set
     * @return boolean - true, if a next row exists, false otherwist
     */
    public boolean next()
    {
      try
      {
        m_currentRow = (Vector)m_rows.elementAt(m_rowNum++);
      }
      catch (Exception e)
      {
        return false;
      }

      return true;
    }

    // getObject method used by the dynamic api to retrive the fields
    // in the current row
    public Object getObject(int pos)
    {
      Object obj = null;

      try
      {
        obj = m_currentRow.elementAt(pos);
      }
      catch (Exception e){}

      return obj;
    }

    /**
     * Add a row to the RowSet's m_rows Vector
     * @param row - A Vector containing the columns
     */
    public void addRow(Vector row)
    {
      m_rows.addElement(row);
    }


    /**
     * Set schema for dynamic temp-table
     * @param schema - the ColumnMetaData[] representing the schema
     */
    public void setSchema(ColumnMetaData[] schema)
    {
      m_schema = schema;
    }

    /**
     * Get schema for dynamic temp-table
     * @return - m_schema, the schema for the RowSet
     */
    public ColumnMetaData[] getSchema()
    {
      return m_schema;
    }

    /**
     * Parse the DataSet element, which contains an <schema> element
     * and a <Data> element.
     * The SchemaParser class parses the <schema> element and turns it
     * into a ColumnMetaData object array (m_schema)
     * The parseData() method in this class parses the <Data> element
     * and builds the m_rows Vector
     * @param dataSetEl - the node containing the schema and data elements
     * @throws Exception
     */
    public void buildRowSet(Node dataSetEl) throws Exception
    {
      // Extract the schema and data elements
      Element schemaEl = null;
      Element dataEl = null;

      NodeList list = dataSetEl.getChildNodes();
      int listLength = list.getLength();
      for (int i = 0; i < listLength; i++)
      {
          Node child = list.item(i);
          if (child.getNodeType() == Node.ELEMENT_NODE)
          {
              String nodeName = child.getLocalName();
              if (schemaEl == null)
              {
                if (!nodeName.equals("schema"))
                  throw new Exception("schema Element missing from dynamic temp-table parameter.");
                schemaEl = (Element) child;
              }
              else if (dataEl == null)
                dataEl = (Element) child;
              else
                throw new Exception("Invalid XML format for dynamic temp-table parameter.");
          }
      }

      // Parse the schema element
      SchemaParser parser = new SchemaParser();
      m_schema = parser.parseSchema(schemaEl);

      if (null == m_schema)
      {
          throw new Exception("Unable to parse schema for dynamic temp-table.");
      }

      // Parse the data element and build up the m_rows Vector
      try{
        parseData(dataEl);
      }
      catch (Exception e)
      {
        throw e;
      }

    }

    /**
     * Parse the xml <Data> element and build up the m_rows vector
     * each element of the m_rows vector is a vector of java objects
     * representing representing the fields in the row
     * @param data - node containing the data for the RowSet
     * @throws Exception
     */
    public void parseData(Node data) throws Exception
    {
      //each child of the data element is a row
      NodeList rowList = data.getChildNodes();
      int rowListLength = rowList.getLength();
      for (int i = 0; i < rowListLength; i++)
      {
        Node rowNode = rowList.item(i);

        if (rowNode.getNodeType() == Node.ELEMENT_NODE)
        {
           Vector row = new Vector();

           // loop through the xml column elements, extracting the values,
           // creating the appropriate Java objects, and adding to the row Vector

           // Need to account for the fact that
           // the temp-table may contain array columns
           NodeList columnList = rowNode.getChildNodes();
           int columnListLength = columnList.getLength();
           int colNum = 0;
           int extentsLeft = 0;
           for (int j = 0; j < columnListLength; j++)
           {
              Node colNode = columnList.item(j);
              if (colNode.getNodeType() == Node.ELEMENT_NODE)
              {
                // Make sure the data doesn't contain too many columns...
                if (colNum >= m_schema.length)
                  throw new Exception ("Too many columns in dynamic temp-table row.");

                String colName = m_schema[colNum].name;
                int colType = m_schema[colNum].type;
                int colExtent = m_schema[colNum].extent;
                if (colExtent > 0 && extentsLeft == 0)
                {
                  extentsLeft = colExtent;
                }


                String value = extractNodeValue(colNode);
                Object valObj = null;

                // get sql type for the element, and create the appropriate
                // object.  Validate the string value.

                // check for "xsi:nil" attribute on the node first
                String nilAttr = ((Element)colNode).getAttribute("xsi:nil");
                if (!nilAttr.equals("true") && !nilAttr.equals("1"))
                {
                  switch(colType)
                  {
                    case ColumnMetaData.CHAR_TYPE:
                      valObj = value;
                    break;
                    case ColumnMetaData.INT_TYPE:
                      try
                      {
                        valObj = Integer.decode(value);
                      }
                      catch (NumberFormatException e)
                      {
                        throw new Exception ("Invalid int value in dynamic temp-table data, " + value);
                      }
                    break;
                    case ColumnMetaData.BOOLEAN_TYPE:
                      if (value.equals("true") || value.equals("1"))
                        valObj =  new Boolean("true");
                      else if (value.equals("false") || value.equals("0"))
                        valObj =  new Boolean("false");
                      else
                        throw new Exception ("Invalid boolean value in dynamic temp-table data, " + value);
                    break;
                    case ColumnMetaData.DECIMAL_TYPE:
                      try
                      {
                        valObj = new BigDecimal(value);
                      }
                      catch (NumberFormatException e)
                      {
                        throw new Exception ("Invalid decimal value in dynamic temp-table data, " + value);
                      }

                    break;
                    case ColumnMetaData.DATE_TYPE:
                      try
                      {
                        Date date=m_sdfDate.parse(value);
                        GregorianCalendar calDate = new GregorianCalendar();
                        calDate.setTime(date);
                        valObj = calDate;
                      }
                      catch (ParseException pe)
                      {
                        throw new Exception ("Invalid date value in dynamic temp-table data, " + value);
                      }
                    break;
                    case ColumnMetaData.RAW_TYPE:
                      valObj = Base64.decode(value);
                    break;
                    case ColumnMetaData.DATETIME_TYPE:
                      valObj = parseDateTime(value);
                    break;
                    default:
                    //shouldn't get here
                    break;
                  }
                }
                row.addElement(valObj);

                // decrement number of array elements left to deal with
                // if extentIndex is zero, advance to next column
                if(extentsLeft == 0 || --extentsLeft == 0)
                  colNum++;
              }
            }

            if (colNum < m_schema.length)
              throw new Exception ("Not enough columns in dynamic temp-table row.");
            // add row to m_rows
            addRow(row);
        }
      }
    }

  /**
   * Get the RowSet as a dom element tree
   * @return org.w3c.dom.Element
   */
   public Element getDataSetAsDom()
   {
      // Create outer "DataSet" element
      Element dataSetEl = m_dom.createElement("DataSet");

      // Add the xmlns="" attribute
      Attr attr = m_dom.createAttribute("xmlns");
      attr.setValue("");
      dataSetEl.setAttributeNode(attr);

      // get the RowSet schema as a DOM Element
      Element schemaEl = getSchemaAsDOM();

      // get the RowSet data as a DOMElement
      Element dataEl = getDataAsDOM();

      // Add schema and data elements as children of the dataset
      dataSetEl.appendChild(schemaEl);
      dataSetEl.appendChild(dataEl);

      return dataSetEl;

   }

    /**
     * Turn m_schema into a dom element tree
     * @return org.w3c.dom.Element, an xml schema
     */
     protected Element getSchemaAsDOM()
     {
        Attr attr;

        // Create the outer schema element
        Element schemaEl = m_dom.createElement("schema");

        attr = m_dom.createAttribute("xmlns");
        attr.setValue("http://www.w3.org/2001/XMLSchema");
        schemaEl.setAttributeNode(attr);

        attr = m_dom.createAttribute("xmlns:xsd");
        attr.setValue("http://www.w3.org/2001/XMLSchema");
        schemaEl.setAttributeNode(attr);

        // Create the Data element and it's children
        Element dataEl = m_dom.createElement("element");

        attr = m_dom.createAttribute("name");
        attr.setValue("Data");
        dataEl.setAttributeNode(attr);

        schemaEl.appendChild(dataEl);

        Element complexTypeEl = m_dom.createElement("complexType");
        dataEl.appendChild(complexTypeEl);

        Element seqEl = m_dom.createElement("sequence");
        complexTypeEl.appendChild(seqEl);

        Element itemEl = m_dom.createElement("element");
        attr = m_dom.createAttribute("name");
        attr.setValue("Item");
        itemEl.setAttributeNode(attr);
        attr = m_dom.createAttribute("minOccurs");
        attr.setValue("0");
        itemEl.setAttributeNode(attr);

        attr = m_dom.createAttribute("maxOccurs");
        attr.setValue("unbounded");
        itemEl.setAttributeNode(attr);

        seqEl.appendChild(itemEl);

        // now for the complexType that contains the metadata for the
        // dynamic temp-table
        Element complexTypeEl2 = m_dom.createElement("complexType");
        itemEl.appendChild(complexTypeEl2);

        Element seqEl2 = m_dom.createElement("sequence");
        complexTypeEl2.appendChild(seqEl2);

        int numCols = m_schema.length;
        for (int i = 0; i < numCols; ++i)
        {
          Element colEl = m_dom.createElement("element");
          attr = m_dom.createAttribute("name");
          attr.setValue(m_schema[i].name);
          colEl.setAttributeNode(attr);
          String typeName;
          switch (m_schema[i].type)
          {
            case ColumnMetaData.CHAR_TYPE:
              typeName = "xsd:string";
              break;
            case ColumnMetaData.INT_TYPE:
              typeName = "xsd:int";
              break;
            case ColumnMetaData.BOOLEAN_TYPE:
              typeName = "xsd:boolean";
              break;
            case ColumnMetaData.DECIMAL_TYPE:
              typeName = "xsd:decimal";
              break;
            case ColumnMetaData.DATE_TYPE:
              typeName = "xsd:date";
              break;
            case ColumnMetaData.RAW_TYPE:
              typeName = "xsd:base64Binary";
              break;
            case ColumnMetaData.DATETIME_TYPE:
              typeName = "xsd:dateTime";
              break;
            default:
              //shouldn't get here
              typeName = "xsd:string";
              break;
          }
          attr = m_dom.createAttribute("type");
          attr.setValue(typeName);
          colEl.setAttributeNode(attr);

          if (m_schema[i].extent > 0)
          {
            attr = m_dom.createAttribute("minOccurs");
            attr.setValue(Integer.toString(m_schema[i].extent));
            colEl.setAttributeNode(attr);

            attr = m_dom.createAttribute("maxOccurs");
            attr.setValue(Integer.toString(m_schema[i].extent));
            colEl.setAttributeNode(attr);
          }

          attr = m_dom.createAttribute("nillable");
          attr.setValue("true");
          colEl.setAttributeNode(attr);

          seqEl2.appendChild(colEl);
        }

        return schemaEl;
     }

     /**
      * Turn the data into a dom element tree
      * @return org.w3c.dom.Element, an xml representation of the data
      */
      protected Element getDataAsDOM()
     {
        int numCols = m_schema.length;

        // Create outer "Data" element
        Element dataEl = m_dom.createElement("Data");

        // now walk through the m_rows Vector, extracting the data
        while (next())
        {
          Element itemEl = m_dom.createElement("Item");
          dataEl.appendChild(itemEl);
          int pos = 0;
          for (int i = 0; i < numCols; i++)
          {
            String colName = m_schema[i].name;
            int colType = m_schema[i].type;
            int colExtent = m_schema[i].extent;

            if (colExtent == 0)
              colExtent = 1;

            for (int j = 0; j < colExtent; j++)
            {
              Object obj = getObject(pos++);

              Element colEl = m_dom.createElement(colName);
              itemEl.appendChild(colEl);

              // Create a Text node child of the column element
              // containing the data

              // Need to specially handle RAW's and DATE's, which don't have
              // a toString() that gives us the text node in a format that we want.
              // Also need to deal with null(UNKNOWN) objects
              if (obj == null)
              {
                Attr attr = m_dom.createAttribute("xsi:nil");
                attr.setValue("true");
                colEl.setAttributeNode(attr);
              }
              else
              {
                switch(m_schema[i].type)
                {
                  case ColumnMetaData.DATE_TYPE:
                    Date calDate = ((GregorianCalendar)obj).getTime();
                    String fdate = m_sdfDate.format(calDate);
                    colEl.appendChild(m_dom.createTextNode(fdate));
                  break;
                  case ColumnMetaData.RAW_TYPE:
                    byte[] bytes = (byte[])obj;
                    colEl.appendChild(m_dom.createTextNode(Base64.encode(bytes)));
                  break;
                  case ColumnMetaData.DATETIME_TYPE:
                    GregorianCalendar dtz = (GregorianCalendar) obj;
                    String fdateTime = getDateTime(dtz);
                    colEl.appendChild(m_dom.createTextNode(fdateTime));
                  break;
                  default:
                    colEl.appendChild(m_dom.createTextNode(obj.toString()));
                  break;
                }
              }
            }
          }
        }
        return dataEl;
     }

     /**
      * Print out the schema and data information to the console
      */
     public void printData()
     {
        int numCols = m_schema.length;
        int rowNum = 1;
        String value;

        // First print out the Schema
        System.out.println("\nRowSet Schema: ");
        System.out.println("colName     colType    colExtent");
        System.out.println("--------------------------------");
        for (int i = 0; i < numCols; i++)
          {
            String colName = m_schema[i].name;
            int colType = m_schema[i].type;
            int colExtent = m_schema[i].extent;

            String typeName;
            switch (m_schema[i].type)
            {
              case ColumnMetaData.CHAR_TYPE:
                typeName = "CHAR       ";
                break;
              case ColumnMetaData.INT_TYPE:
                typeName = "INT        ";
                break;
              case ColumnMetaData.BOOLEAN_TYPE:
                typeName = "LOGICAL    ";
                break;
              case ColumnMetaData.DECIMAL_TYPE:
                typeName = "DECIMAL    ";
                break;
              case ColumnMetaData.DATE_TYPE:
                typeName = "DATE       ";
                break;
              case ColumnMetaData.RAW_TYPE:
                typeName = "RAW        ";
                break;
              case ColumnMetaData.DATETIME_TYPE:
                typeName = "DATETIME   ";
                break;
              default:
                //shouldn't get here
                typeName = "UNKNOWN    ";
                break;
            }

		// Line up the strings
		int nameFill = 0;
		if (colName.length() < 12)
			nameFill = 12 - colName.length();

		System.out.print(colName);

		for (int j = 0; j < nameFill; ++j)
			System.out.print(" ");

		System.out.println(typeName + "  "+ colExtent);
          }

        System.out.println("\nRowSet Data: ");
        System.out.println("------------------------------");
        // walk through the m_rows Vector, extracting the data
        while (next())
        {
          //System.out.println("\nRow number: " + rowNum++);
          int pos = 0;
          for (int i = 0; i < numCols; i++)
          {
            String colName = m_schema[i].name;
            int colType = m_schema[i].type;
            int colExtent = m_schema[i].extent;

            if (colExtent == 0)
              colExtent = 1;

            for (int j = 0; j < colExtent; j++)
            {
              Object obj = getObject(pos++);

              if (obj == null)
              {
                  System.out.print("null    ");
              }
              else
              {
                switch(m_schema[i].type)
                {
                  case ColumnMetaData.DATE_TYPE:
                    Date calDate = ((GregorianCalendar)obj).getTime();
                    value = m_sdfDate.format(calDate);
                  break;
                  case ColumnMetaData.RAW_TYPE:
                    byte[] bytes = (byte[])obj;
                    value = Base64.encode(bytes);
                  break;
                  case ColumnMetaData.DATETIME_TYPE:
                    GregorianCalendar dtz = (GregorianCalendar) obj;
                    value = getDateTime(dtz);
                  break;
                  default:
                    value = obj.toString();
                  break;
                }
                System.out.print(value + "    ");
              }
            }
          }
          System.out.println();
        }
        System.out.println();
     }

    /**
     * Extracts a child text node from an element node and returns the value as a string. This
     * method assumes a single child node of type TEXT. If there aren't any child nodes or the
     * first child is not a text node, the method returns an empty string.
     *
     * @param   n       An org.w3c.dom.Node that represents an ELEMENT node and has children.
     * @return  String  Contains the value of a child TEXT node of the node passed in, or
     *                  an empty string.
     */
    public static String extractNodeValue(Node n)
    {
        String retVal = "";
        if ((n.getNodeType() == Node.ELEMENT_NODE) && n.hasChildNodes())
        {
            Node text = n.getFirstChild();
            if (text.getNodeType() == Node.TEXT_NODE)
                retVal = text.getNodeValue();
        }
        return retVal;

    }
    
    /**
     * Parses the xml dateTime element and creates a GregorianCalendar
     * object with the proper date and time set.
     * 
     * @param  orgvalue    - xml string containing the data for the dateTime
     * @return dateTimeVal - a GregorianCalendar containing the dateTime
     * @throws Exception
     **/
    private GregorianCalendar parseDateTime(String orgvalue) throws Exception
    {
        GregorianCalendar dateTimeVal = null;
        try
        {
          /* The format of the dateTime is *REQUIRED* to include
           *        YYYY-MM-DDThh:mm:ssZ
           * at least.
           * Can be optionally preceded by "-"
           * Can also include optional miliseconds ss.nnn (any # of digits)
           * Time zone can be represented with "Z" at the end for UTC
           *     or
           *     +/-hh:mm for a time zone offset.
           */
          int negativeDatetime = 0;
          if(orgvalue.substring(0) == "-")
              negativeDatetime = 1;
          String value = new String(orgvalue.substring(negativeDatetime, orgvalue.length()));

          if (value.length() <= 19)
          {
            // cannot be valid format if shorter than 20 characters
            throw new Exception ("Invalid dateTime value in dynamic temp-table data, " + value);
          }

          /* Need to deal with all formats mentioned above
           * The easiest way to get the miliseconds is to do it based on how
           * the string ends. So set up for the miliseconds and then do it
           * based upon which time zone type we have.
           */
          String timezone = new String("GMT");
          int milisec = 0;
          int decimalpos = value.indexOf(".") + 1;
          // if decimalpos > 0 then have milisecond
          int finalpos;
                            
          if(value.indexOf("Z") != -1)
          // Then it is UTC
          {
            timezone = timezone.concat("+00:00");
            finalpos = value.indexOf("Z");
          }
          else if(value.indexOf("+") != -1)
          // Then it is positive off GMT
          {
            timezone = timezone.concat(value.substring(value.length() - 6, value.length()));
            finalpos = value.indexOf("+");
          }
          else if(value.indexOf("-", value.indexOf("T")) != -1)
          // Then it is negative off GMT
          {
            timezone = timezone.concat(value.substring(value.length() - 6, value.length()));
            finalpos = value.lastIndexOf("-");
          }
          else
          // no time zone information
          // This is not allowed - throw an exception
          {
            throw new Exception ("Invalid dateTime value in dynamic temp-table data, " + value);
          }

          if(decimalpos != 0)
          {
            int diff = finalpos-decimalpos;
            if(diff >= 3)
            {
              // reuse diff for rounding to three places
              finalpos = decimalpos+3;
              if(diff > 3)
              {
                Integer rounder = new Integer(value.substring(decimalpos+4, decimalpos+5));
                if(rounder.intValue() >= 5) diff = 1;
                else diff = 0;
              }
              else diff = 0;
            }
            Integer milsec = new Integer(value.substring(decimalpos, finalpos));
            milisec = milsec.intValue() + diff;
          }

          Date date = m_sdfDateTime.parse(value.substring(0,19)+timezone);
          dateTimeVal = new GregorianCalendar();
          dateTimeVal.setTime(date);
          dateTimeVal.add(java.util.Calendar.MILLISECOND, milisec);
          dateTimeVal.setTimeZone(TimeZone.getTimeZone(timezone));
        }
        catch (ParseException pe)
        {
          throw new Exception ("Invalid dateTime value in dynamic temp-table data, " + orgvalue);
        }
        return dateTimeVal;
    }
    
    /**
     * Takes a GregorianCalendar object and generates a string that is
     * a proper xml dateTime using the date and time from the input.
     * 
     * @param  dtz       - a GregorianCalendar to be turned into xml
     * @return fdateTime - a String whose data is the xml Schema dateTime data
     * @throws Exception
     **/
    private String getDateTime(GregorianCalendar dtz)
    {
        DecimalFormat yearDigits  = new DecimalFormat("#0000;-#0000");
        DecimalFormat twoDigits   = new DecimalFormat("00");
        DecimalFormat threeDigits = new DecimalFormat("000");
        StringBuffer  fdateTime   = new StringBuffer();
        int iVal;

        iVal = dtz.get(Calendar.YEAR);
        fdateTime.append(yearDigits.format(iVal) + "-");
        iVal = dtz.get(Calendar.MONTH);
        fdateTime.append(twoDigits.format(iVal + 1) + "-");
        iVal = dtz.get(Calendar.DAY_OF_MONTH);
        fdateTime.append(twoDigits.format(iVal) + "T");
        iVal = dtz.get(Calendar.HOUR_OF_DAY);
        fdateTime.append(twoDigits.format(iVal) + ":");
        iVal = dtz.get(Calendar.MINUTE);
        fdateTime.append(twoDigits.format(iVal) + ":");
        iVal = dtz.get(Calendar.SECOND);
        fdateTime.append(twoDigits.format(iVal) + ".");
        iVal = dtz.get(Calendar.MILLISECOND);
        fdateTime.append(threeDigits.format(iVal));
        // account for daylight savings in the timezone
        int tzOff = dtz.get(Calendar.ZONE_OFFSET);
        int dstOff = dtz.get(Calendar.DST_OFFSET);
        iVal = (tzOff + dstOff) / 1000;
        if (iVal < 0)
        {
            fdateTime.append("-");
          iVal *= -1;
        }
        else fdateTime.append("+");
        fdateTime.append(twoDigits.format(iVal / 3600) + ":");
        fdateTime.append(twoDigits.format(((iVal / 60) % 60)) );
        return fdateTime.toString();
    }
}

