/*
 * Class:    SchemaParser
 *
 */

import  java.lang.*;
import java.util.*;
import org.w3c.dom.*;


/**
 * Class to parse an xml schema for a dynamic temp-table output parameter
 * Recursively walks the dom tree for the schema
 * looking for a complexType containing the sequence of elements that define the
 * dynamic temp-table's schema
 */
public class SchemaParser
{

    protected ColumnMetaData[] m_schema;

    /**
    * <!-- () -->
    * <p>The default class constructor.
    * </p>
    * <br>
    *return   void
    */
    public SchemaParser()
    {
      m_schema = null;
    }

    /**
     * Parses "Data" element declaration in the schema
     * @param   n   An org.w3c.dom.Node that represents the <schema> element.
     * @return  ColumnMetaData[]  the metadata array representing the schema
     */
    /* Parses "Data" element declaration in the schema*/
    public ColumnMetaData[] parseSchema(Node n)
    {
        NodeList list;
        int listLength, i;
        Node child, dataNode = null;

        // "Data" element is first child of schema node
        list = n.getChildNodes();
        listLength = list.getLength();
        for (i = 0; i < listLength; i++)
        {
            child = list.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE)
            {
                dataNode = child;
                break;
            }

        }

        if (dataNode != null)
        {
          list = dataNode.getChildNodes();
          listLength = list.getLength();
          for (i = 0; i < listLength; i++)
          {
              child = list.item(i);
              if (child.getNodeType() == Node.ELEMENT_NODE)
              {
                  String nodeName = child.getLocalName();
                  if (nodeName.equals("complexType"))
                  {
                        parseComplexType(child);
                  }
              }

          }
        }

        return m_schema;
    }

    /**
     * Parse a complexType node
     * @param   n   An org.w3c.dom.Node containg the <complexType> element.
     */
    void parseComplexType(Node n)
    {

        /* A complexType will have a sequence, choice, or all child */
        NodeList list = n.getChildNodes();
        int listLength = list.getLength();
        for (int i = 0; i < listLength; i++)
        {
            Node child = list.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE)
            {
                String nodeName = child.getLocalName();
                if (nodeName.equals("sequence"))
                {
                    /* Check if sequence contains only elements of
                    ** simple types (no children).  If so, create schema
                    ** based on this sequence. Else, continue parsing.
                    */
                    int numCols = checkSequence(child);
                    if (numCols != -1)
                    {
                      createSchema(child, numCols);
                      return;
                    }

                    /* This sequence is not our temp-table definition
                    ** Continue parsing
                    */
                    NodeList sequence = child.getChildNodes();
                    int seqLength = sequence.getLength();
                    for (int j = 0; j < seqLength; j++)
                    {
                        Node seqChild = sequence.item(j);
                        if (seqChild.getNodeType() == Node.ELEMENT_NODE)
                        {
                            /* Make sure child is "element" */
                            String seqName = seqChild.getLocalName();
                            if (seqName.equals("element"))
                            {
                                parseElement(seqChild);
                            }
                        }
                    }
                }
                else if (nodeName.equals("choice"))
                {
                    /* A choice can have one or more elements */
                    NodeList choice = child.getChildNodes();
                    int choiceLength = choice.getLength();
                    for (int j = 0; j < choiceLength; j++)
                    {
                        Node choiceChild = choice.item(j);
                        if (choiceChild.getNodeType() == Node.ELEMENT_NODE)
                        {
                            /* Make sure child is "element" */
                            String choiceName = choiceChild.getLocalName();
                            if (choiceName.equals("element"))
                            {
                                parseElement(choiceChild);
                            }
                            /* Continue in loop. No need to throw excpetion at this point */
                        }
                    }
                }
                else if (nodeName.equals("all"))
                {
                    // Throw Exception. Don't even attempt
                }
            }
        }
    }

    /**
     * Parse an <element> node, looking for a <complexType>
     * @param   n An org.w3c.dom.Node that contains an <element> element.
     */
    void parseElement(Node n)
    {

        /* An element will have a complexType, or have a named simple type */
        /* We only care about getting to the complexType definition for the */
        /* sequence of temp-table fields */
        NodeList list = n.getChildNodes();
        int listLength = list.getLength();
        for (int i = 0; i < listLength; i++)
        {
            Node child = list.item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE)
            {
                String nodeName = child.getLocalName();
                if (nodeName.equals("complexType"))
                    parseComplexType(child);
            }
        }


    }

    /**
     * Parse a sequence node, checking whether the sequence contains
     * only elements of simple type (no child nodes).
     * @param   seqNode   An org.w3c.dom.Node containing a <sequence> element.
     * @return  int       The number of element nodes in the sequence, -1 if the
     *                    sequence has elements with child nodes.
     */
   int checkSequence(Node seqNode)
   {
        int numElements = 0;
        NodeList sequence = seqNode.getChildNodes();
        int seqLength = sequence.getLength();
        for (int j = 0; j < seqLength; j++)
        {
        Node seqChild = sequence.item(j);
        if (seqChild.getNodeType() == Node.ELEMENT_NODE)
        {
          /* Make sure child is "element" */
          String seqName = seqChild.getLocalName();
          if (!seqName.equals("element"))
          {
            return -1;
          }

          /* make sure element has no children */
          Node child = seqChild.getFirstChild();
          if (child != null)
            return -1;
          else
            ++numElements;
        }

      }

      return numElements;
   }

   /**
   * Create the schema vector  based on the sequence passed in
   *
   */

   /**
    * Create the m_schema ColumnMetaData array  based on the sequence passed in
    * @param  child   An org.w3c.dom.Node containing a <sequence> element
    * @param  numCols number of child elements of the 'child' node
    */
   void createSchema (Node child, int numCols)
   {

      NodeList sequence = child.getChildNodes();
      Node seqChild;
      int seqLength = sequence.getLength();
      String colName, xmlType, colMaxOccurs;
      int index, colType, colNum, extent;

      m_schema = new ColumnMetaData[numCols];

      // Parse the sequence, extracting the columns name and type
      // Create a ColumnSchema object and add it to the m_schema vector
      colNum = 0;
      for (int j = 0; j < seqLength; j++)
      {
          seqChild = sequence.item(j);
          if (seqChild.getNodeType() == Node.ELEMENT_NODE)
          {
            // get column name
            colName = ((Element)seqChild).getAttribute("name");
            if (colName.length() == 0)
            {
              m_schema = null;
              return;
            }

            // get column type
            xmlType = ((Element)seqChild).getAttribute("type");
            // get local Part of the type
            index = xmlType.indexOf(':');
            if (index != -1)
              xmlType = xmlType.substring(index + 1);

            // translate xml type to ColumnMetaData static type, and add to the Vector
            // We don't support all of the scalar datatypes for dynamic temp-tables
            if (xmlType.equals("string"))
              colType = ColumnMetaData.CHAR_TYPE;
            else if (xmlType.equals("int"))
              colType = ColumnMetaData.INT_TYPE;
            else if (xmlType.equals("boolean"))
              colType = ColumnMetaData.BOOLEAN_TYPE;
            else if (xmlType.equals("decimal"))
              colType = ColumnMetaData.DECIMAL_TYPE;
            else if (xmlType.equals("date"))
              colType = ColumnMetaData.DATE_TYPE;
            else if (xmlType.equals("base64Binary"))
              colType = ColumnMetaData.RAW_TYPE;
            else if (xmlType.equals("dateTime"))
              colType = ColumnMetaData.DATETIME_TYPE;
            else
            {
              m_schema = null;
              return;
            }

            // get the column extent, maxOccurs attribute
            colMaxOccurs = ((Element)seqChild).getAttribute("maxOccurs");
            extent = 0;
            if (colMaxOccurs.length() != 0)
            {
              try
              {
                Integer iExt = Integer.decode(colMaxOccurs);
                extent = iExt.intValue();
                if (extent <= 1)
                  extent = 0;
              }
              catch (NumberFormatException e)
              {
                m_schema = null;
                return;
              }
            }

            // create the ColumnMetaData object and add to the m_schema array
            ColumnMetaData md = new ColumnMetaData(colName, colType, extent);
            m_schema[colNum] = md;
            colNum++;

          }
        }
   }
}

