/*
 * SOAPHeader2.p
 *
 * An addition to SOAPHeader1.p.
 *
 * Calls a ficticious Web service.
 * The first operation 'OpenAccess' sends nothing in the request headers and 
 * gets back a SOAP response header containing an AccessID.
 * The second operation 'HelloWorld' sends the AccessID back its request header
 * no additional information is received in the response header.
 * The third operation 'HelloSecureWorld' adds a Password node to the 
 * existing AccessID entry in its request header. This Password node is added
 * as a sibling of the AccessID element and NOT as a new SOAP header entry. Once
 * again no additional information is received in the response header.
 *
 * The Web service has only one service and port available.
 *
 */

/* Define local variables */
DEF VAR hWebSrvc      AS HANDLE.
DEF VAR hPortType     AS HANDLE.
DEF VAR cResponse     AS CHAR FORMAT "x(72)".
DEF VAR hXdoc         AS HANDLE.
DEF VAR hXnoderef1    AS HANDLE.
DEF VAR hXnoderef2    AS HANDLE.
DEF VAR hXtext        AS HANDLE.
DEF VAR lDeleteOnDone AS LOGICAL.
DEF VAR cPassword     AS CHARacter INIT ?.
DEF VAR g_header      AS HANDLE.

/* Create the Web service server object */
CREATE SERVER hWebSrvc.

/* Connect to the WS */
hWebSrvc:CONNECT("-WSDL http://ServiceHost/SOAPHeader/HeaderExample.asmx?wsdl").

/* Get the method, set the port type */
RUN HeadersSoap SET hPortType ON hWebSrvc.

/* Associate the req. & resp. callbacks with the port type */
hPortType:SET-CALLBACK-PROCEDURE("REQUEST-HEADER", "ReqHandler").
hPortType:SET-CALLBACK-PROCEDURE("RESPONSE-HEADER", "RespHandler").

/* Invoke the Web service with no header and display the results */
RUN OpenAccess IN hPortType (OUTPUT cResponse).
DISP cResponse LABEL "WS response" WITH FRAME aaa.

/* Go again with the AccessID set from previous response */
cResponse = "".
RUN HelloWorld IN hPortType (OUTPUT cResponse).
DISP cResponse LABEL "WS response" WITH FRAME bbb.

/* Go again with the AccessID set from previous response */
/* header together with an added Username and Password   */
cResponse = "".
cPassword = "Administrator".
RUN HelloSecureWorld IN hPortType (OUTPUT cResponse).
DISP cResponse LABEL "WS response" WITH FRAME ccc.

DELETE OBJECT g_header.
DELETE OBJECT hPortType.
hWebSrvc:DISCONNECT().
DELETE OBJECT hWebSrvc.

/**************** Internal Procedures ****************/
PROCEDURE ReqHandler:
    DEFINE OUTPUT PARAMETER hHeader    AS HANDLE.
    DEFINE INPUT  PARAMETER cNamespace AS CHAR.
    DEFINE INPUT  PARAMETER cLocalNS   AS CHAR.
    DEFINE OUTPUT PARAMETER lDeleteOnDone AS LOGICAL.

        /* The IF test determines if this is the first call through this code.
           If it is, then g_header is not set and hHeader is set
           to ? to ensure that no header is sent.
           g_header gets set when the response header is
           returned, so a subsequent pass through this code takes
           the previous response header and sends it as the
           current request header, possibly modified to 
           authenticate a secure request. */
    IF NOT VALID-HANDLE (g_header) THEN DO: /* first request */
        hHeader = ?.
        lDeleteOnDone = TRUE.
    END.
    ELSE DO: /* all subsequent requests */
        hHeader = g_header.
        lDeleteOnDone = FALSE.

        /* Password node data are added to the existing SOAP header 
           if a secure operation is being executed */
        IF cPassword <> ? THEN DO:
            /* Build a new x-doc to contain the Password data */
            CREATE X-DOCUMENT hXdoc.
            CREATE X-NODEREF  hXnoderef1.
            CREATE X-NODEREF  hXnoderef2.
            CREATE X-NODEREF  hXtext.

            /* Add the Password data as a child of the existing header entry */
            DEFINE VARIABLE hHeaderEntryref AS HANDLE.
            CREATE SOAP-HEADER-ENTRYREF hHeaderEntryref.

            hHeader:GET-HEADER-ENTRY(hHeaderEntryref, 1).
            hHeaderEntryref:GET-NODE(hXnoderef1).

            hXdoc:IMPORT-NODE(hXnoderef2, hXnoderef1, TRUE).
            hXdoc:APPEND-CHILD(hXnoderef2).
            hXdoc:CREATE-NODE-NAMESPACE(hXnoderef1, hHeaderEntryref:NAMESPACE-URI, "Password", "ELEMENT").
            hXnoderef2:APPEND-CHILD(hXnoderef1).
            hXdoc:CREATE-NODE(hXtext, "", "text").
            hXnoderef1:APPEND-CHILD(hXtext).
            hXtext:NODE-VALUE = cPassword.

            /* Replace the existing header entry using a deep copy of the new */
            /* version from the x-doc that has the Password node added        */
            hHeaderEntryref:SET-NODE(hXnoderef2).

            /* Procedure/header cleanup */
            cPassword = ?.
            DELETE OBJECT hHeaderEntryref.
            DELETE OBJECT hXdoc.
            DELETE OBJECT hXnoderef1.
            DELETE OBJECT hXnoderef2.
            DELETE OBJECT hXtext.
        END.
    END.
END PROCEDURE.

PROCEDURE RespHandler:
    DEFINE INPUT PARAMETER hHeader    AS HANDLE.
    DEFINE INPUT PARAMETER cNamespace AS CHAR.
    DEFINE INPUT PARAMETER cLocalNS   AS CHAR.
        
        /* If the g_header global variable 
           is valid coming in, it has already been set 
           in a previous response, therefore, delete 
           the unnecessary response header object.
           Otherwise, set g-header to the response 
           header object to pass back to the request 
           header handler on subsequent requests. */
    IF VALID-HANDLE(g_header) THEN
        DELETE OBJECT hHeader.
    ELSE DO:
        g_header = hHeader.
    END.    
END PROCEDURE.
